@@ -4,23 +4,26 @@ Uses IPython parallel to setup a cluster and manage execution:
 
 http://ipython.org/ipython-doc/stable/parallel/index.html
 Borrowed from Brad Chapman's implementation:
-https://github.com/chapmanb/bcbio-nextgen/blob/master/bcbio/pipeline/ipython.py
+https://github.com/chapmanb/bcbio-nextgen/blob/master/bcbio/distributed/ipython.py
 """
-import pipes
-import time
-import uuid
-import subprocess
 import contextlib
+import os
+import pipes
+import uuid
+import shutil
+import subprocess
+import time
+
 from IPython.parallel import Client
 from IPython.parallel.apps import launcher
 from IPython.utils import traitlets
-import os
-import shutil
+from IPython.utils.traitlets import (List, Unicode, CRegExp)
 
 # ## Custom launchers
 
-timeout_params = ["--timeout=30", "--IPEngineApp.wait_for_url_file=120"]
+timeout_params = ["--timeout=30", "--IPEngineApp.wait_for_url_file=960"]
 
+# ## Platform LSF
 class BcbioLSFEngineSetLauncher(launcher.LSFEngineSetLauncher):
     """Custom launcher handling heterogeneous clusters on LSF.
     """
@@ -43,11 +46,12 @@ class BcbioLSFControllerLauncher(launcher.LSFControllerLauncher):
     default_template = traitlets.Unicode("""#!/bin/sh
 #BSUB -J bcbio-ipcontroller
 #BSUB -oo bcbio-ipcontroller.bsub.%%J
-%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
+%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --nodb --hwm=5 --scheme=pure
     """%(' '.join(map(pipes.quote, launcher.ipcontroller_cmd_argv))))
     def start(self):
         return super(BcbioLSFControllerLauncher, self).start()
 
+# ## Sun Grid Engine (SGE)
 class BcbioSGEEngineSetLauncher(launcher.SGEEngineSetLauncher):
     """Custom launcher handling heterogeneous clusters on SGE.
     """
@@ -75,7 +79,7 @@ class BcbioSGEControllerLauncher(launcher.SGEControllerLauncher):
     default_template = traitlets.Unicode(u"""#$ -V
 #$ -S /bin/sh
 #$ -N ipcontroller
-%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
+%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --nodb --hwm=5 --scheme=pure
 """%(' '.join(map(pipes.quote, launcher.ipcontroller_cmd_argv))))
     def start(self):
         return super(BcbioSGEControllerLauncher, self).start()
@@ -92,14 +96,14 @@ def _find_parallel_environment():
                      "See %s for SGE setup instructions." %
                      "https://blogs.oracle.com/templedf/entry/configuring_a_new_parallel_environment")
 
+# ## PBS
 class BcbioPBSEngineSetLauncher(launcher.PBSEngineSetLauncher):
     """Custom launcher handling heterogeneous clusters on SGE.
     """
     cores = traitlets.Integer(1, config=True)
     pename = traitlets.Unicode("", config=True)
     default_template = traitlets.Unicode("""#PBS -V
-#PBS -cwd
-#PBS -j y
+#PBS -j oe
 #PBS -S /bin/sh
 #PBS -q {queue}
 #PBS -N bcbio-ipengine
@@ -111,28 +115,101 @@ class BcbioPBSEngineSetLauncher(launcher.PBSEngineSetLauncher):
     def start(self, n):
         self.context["cores"] = self.cores
         self.context["pename"] = str(self.pename)
-        return super(BcbioSGEEngineSetLauncher, self).start(n)
+        return super(BcbioPBSEngineSetLauncher, self).start(n)
 
 
 class BcbioPBSControllerLauncher(launcher.PBSControllerLauncher):
     default_template = traitlets.Unicode(u"""#PBS -V
 #PBS -S /bin/sh
 #PBS -N ipcontroller
-%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
+%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --nodb --hwm=5 --scheme=pure
 """%(' '.join(map(pipes.quote, launcher.ipcontroller_cmd_argv))))
 
     def start(self):
         return super(BcbioPBSControllerLauncher, self).start()
 
-class BcbioTORQUEEngineSetLauncher(BcbioPBSEngineSetLauncher):
-    def start(self):
-        return super(BcbioTORQUEEngineSetLauncher, self).start()
+# ## Torque
+class TORQUELauncher(launcher.BatchSystemLauncher):
+    """A BatchSystemLauncher subclass for PBS."""
 
-class BcbioTORQUEControllerLauncher(BcbioPBSControllerLauncher):
-    def start(self):
-        return super(BcbioTORQUEControllerLauncher, self).start()
+    submit_command = List(['qsub'], config=True,
+        help="The PBS submit command ['qsub']")
+    delete_command = List(['qdel'], config=True,
+        help="The PBS delete command ['qsub']")
+    job_id_regexp = CRegExp(r'\d+', config=True,
+        help="Regular expresion for identifying the job ID [r'\d+']")
 
-# ## Control clusters
+    batch_file = Unicode(u'')
+    job_array_regexp = CRegExp('#PBS\W+-t\W+[\w\d\-\$]+')
+    job_array_template = Unicode('#PBS -t 1-{n}')
+    queue_regexp = CRegExp('#PBS\W+-q\W+\$?\w+')
+    queue_template = Unicode('#PBS -q {queue}')
+
+class BcbioTORQUEEngineSetLauncher(TORQUELauncher, launcher.BatchClusterAppMixin):
+    """Launch Engines using PBS"""
+    batch_file_name = Unicode(u'torque_engines', config=True,
+        help="batch file name for the engine(s) job.")
+    default_template= Unicode(u"""#!/bin/sh
+#PBS -V
+#PBS -N ipengine
+%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
+"""%(' '.join(map(pipes.quote,launcher.ipengine_cmd_argv))))
+
+    def start(self, n):
+        """Start n engines by profile or profile_dir."""
+        return super(BcbioTORQUEEngineSetLauncher, self).start(n)
+
+class BcbioTORQUEControllerLauncher(TORQUELauncher, launcher.BatchClusterAppMixin):
+    """Launch a controller using PBS."""
+
+    batch_file_name = Unicode(u'torque_controller', config=True,
+        help="batch file name for the controller job.")
+    default_template= Unicode("""#!/bin/sh
+#PBS -V
+#PBS -N ipcontroller
+%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --nodb --hwm=5 --scheme=pure
+"""%(' '.join(map(pipes.quote, launcher.ipcontroller_cmd_argv))))
+
+
+    def start(self):
+        """Start the controller by profile or profile_dir."""
+        return super(BcbioTORQUEControllerLauncher, self).start(1)
+
+# ## PBSPro
+class PBSPROLauncher(launcher.PBSLauncher):
+    """A BatchSystemLauncher subclass for PBSPro."""
+    job_array_regexp = CRegExp('#PBS\W+-J\W+[\w\d\-\$]+')
+    job_array_template = Unicode('#PBS -J 1-{n}')
+
+class BcbioPBSPROEngineSetLauncher(PBSPROLauncher, launcher.BatchClusterAppMixin):
+    """Launch Engines using PBSPro"""
+    batch_file_name = Unicode(u'pbspro_engines', config=True,
+        help="batch file name for the engine(s) job.")
+    default_template= Unicode(u"""#!/bin/sh
+#PBS -V
+#PBS -N ipengine
+%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
+"""%(' '.join(map(pipes.quote,launcher.ipengine_cmd_argv))))
+
+    def start(self, n):
+        """Start n engines by profile or profile_dir."""
+        return super(BcbioPBSPROEngineSetLauncher, self).start(n)
+
+class BcbioPBSPROControllerLauncher(PBSPROLauncher, launcher.BatchClusterAppMixin):
+    """Launch a controller using PBSPro."""
+
+    batch_file_name = Unicode(u'pbspro_controller', config=True,
+        help="batch file name for the controller job.")
+    default_template= Unicode("""#!/bin/sh
+#PBS -V
+#PBS -N ipcontroller
+%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --nodb --hwm=5 --scheme=pure
+"""%(' '.join(map(pipes.quote, launcher.ipcontroller_cmd_argv))))
+
+
+    def start(self):
+        """Start the controller by profile or profile_dir."""
+        return super(BcbioPBSPROControllerLauncher, self).start(1)
 
 def _start(scheduler, profile, queue, num_jobs, cores_per_job):
     """Starts cluster from commandline.
@@ -144,18 +221,19 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job):
     args = launcher.ipcluster_cmd_argv + \
         ["start",
          "--daemonize=True",
-         "--IPClusterEngines.early_shutdown=180",
-         "--delay=20",
+         "--IPClusterEngines.early_shutdown=240",
+         "--delay=10",
          "--log-level=%s" % "WARN",
          "--profile=%s" % profile,
          "--n=%s" % num_jobs,
          "--%s.cores=%s" % (engine_class, cores_per_job),
          "--IPClusterStart.controller_launcher_class=%s.%s" % (ns, controller_class),
          "--IPClusterStart.engine_launcher_class=%s.%s" % (ns, engine_class),
-         "--%sLauncher.queue='%s'" % (scheduler, queue),
+         "--%sLauncher.queue='%s'" % (scheduler, queue)
          ]
     if scheduler in ["SGE"]:
         args += ["--%s.pename=%s" % (engine_class, _find_parallel_environment())]
+
     subprocess.check_call(args)
 
 def _stop(profile):
@@ -173,7 +251,7 @@ def _is_up(profile, n):
         return up >= n
 
 @contextlib.contextmanager
-def cluster_view(scheduler, queue, num_jobs, cores_per_job=1):
+def cluster_view(scheduler, queue, num_jobs, cores_per_job=1, profile=None):
     """Provide a view on an ipython cluster for processing.
 
     parallel is a dictionary with:
@@ -181,10 +259,14 @@ def cluster_view(scheduler, queue, num_jobs, cores_per_job=1):
       - num_jobs: Number of jobs to start.
       - cores_per_job: The number of cores to use for each job.
     """
-    delay = 5
-    max_delay = 300
+    delay = 10
+    max_delay = 960
     max_tries = 10
-    profile = create_throwaway_profile()
+    if profile is None:
+        has_throwaway = True
+        profile = create_throwaway_profile()
+    else:
+        has_throwaway = False
     num_tries = 0
     while 1:
         try:
@@ -215,13 +297,16 @@ def cluster_view(scheduler, queue, num_jobs, cores_per_job=1):
         if client:
             client.close()
         _stop(profile)
-        delete_profile(profile)
+        if has_throwaway:
+            delete_profile(profile)
 
 def _get_balanced_blocked_view(client):
     view = client.load_balanced_view()
     view.block = True
     return view
 
+# ## Temporary profile management
+
 def create_throwaway_profile():
     profile = str(uuid.uuid1())
     cmd = "ipython profile create {0} --parallel".format(profile)
